'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = createAction;
exports.handleActionSuccess = handleActionSuccess;
exports.handleActionBusy = handleActionBusy;
exports.handleActionError = handleActionError;
exports.handleActionCancellation = handleActionCancellation;
exports.errorType = errorType;
exports.successType = successType;
exports.busyType = busyType;
exports.cancellationType = cancellationType;

var _getActionName = require('./getActionName');

var _getActionName2 = _interopRequireDefault(_getActionName);

var _types = require('./types');

var _next = require('./next');

var _config = require('./config');

var _constants = require('./constants');

var _listeners = require('./listeners');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function createAction(action) {
  if (typeof action !== 'function') throw new Error('Action must be a function');

  if (action._trackedAction) return action._trackedAction;

  var trackedAction = function trackedAction() {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    var store = (0, _config.getStore)();
    var handleError = handleActionError.bind(null, action, args);
    var extras = {};
    var actionContext = { setExtra: function setExtra(values) {
        return Object.assign(extras, values);
      } };

    try {

      handleActionBusy(action, args);

      var actionResponse = action.apply(actionContext, [].concat(args, [store]));
      var isAsync = actionResponse instanceof Promise;

      if (isAsync) {
        var cancelled = false;
        var cancellable = true;

        var cancellableResponse = actionResponse.then(function (payload) {
          cancellable = false;
          return !cancelled ? handleActionSuccess(action, payload, args) : null;
        }, function (err) {
          cancellable = false;
          return !cancelled ? Promise.reject(handleError(err)) : null;
        });

        cancellableResponse.cancel = function () {
          if (!cancellable) console.warn('\n              Action (' + actionName + ') is not executing anymore, therefore can\'t be cancelled.\n              Check \'canBeCancelled\' prop to verify whether an action call can be cancelled or not.');

          cancelled = true;
          cancellable = false;

          handleActionCancellation(action, args, extras);

          return { args: args, extras: extras };
        };

        Object.defineProperty(cancellableResponse, 'canBeCancelled', {
          enumerable: true,
          get: function get() {
            return cancellable;
          }
        });

        return cancellableResponse;
      }

      return handleActionSuccess(action, actionResponse, args);
    } catch (err) {

      handleError(err);
      throw err;
    }
  };

  var actionName = (0, _getActionName2.default)(action._prefix || (0, _config.getDefaultPrefix)(), action.name, action);

  trackedAction._successType = actionName;
  trackedAction._busyType = actionName + (0, _config.getBusySuffix)();
  trackedAction._errorType = actionName + (0, _config.getErrorSuffix)();
  trackedAction._cancellationType = actionName + (0, _config.getCancelSuffix)();

  action._trackedAction = trackedAction;

  return trackedAction;
}

function handleActionSuccess(action, payload, args) {
  var store = (0, _config.getStore)();

  store.dispatch({
    _actionwareType: _constants.SUCCESS_TYPE,
    _trackedAction: action._trackedAction,
    type: action._trackedAction._successType,
    payload: payload
  });

  (0, _listeners.notifySuccessListeners)({ action: action, payload: payload, args: args });
  (0, _next.resolveWaiters)(action, payload);

  return payload;
}

function handleActionBusy(action, args) {
  var store = (0, _config.getStore)();

  store.dispatch({
    _actionwareType: _constants.BUSY_TYPE,
    _trackedAction: action._trackedAction,
    type: action._trackedAction._busyType,
    payload: null, // Payload prop must be set
    args: args
  });

  (0, _listeners.notifyBeforeListeners)({ action: action, args: args });
}

function handleActionError(action, args, error) {
  var store = (0, _config.getStore)();

  store.dispatch({
    _actionwareType: _constants.ERROR_TYPE,
    _trackedAction: action._trackedAction,
    type: action._trackedAction._errorType,
    payload: error,
    args: args
  });

  (0, _listeners.notifyErrorListeners)({ action: action, error: error, args: args });
  (0, _next.rejectWaiters)(action, error);

  return error;
}

function handleActionCancellation(action, args, extras) {
  var store = (0, _config.getStore)();

  store.dispatch({
    _actionwareType: _constants.CANCEL_TYPE,
    _trackedAction: action._trackedAction,
    type: action._trackedAction._cancellationType,
    payload: extras,
    args: args
  });

  (0, _listeners.notifyCancellationListeners)({ action: action, args: args, extras: extras });
}

function errorType(action) {
  return createAction(action)._errorType;
}

function successType(action) {
  return createAction(action)._successType;
}

function busyType(action) {
  return createAction(action)._busyType;
}

function cancellationType(action) {
  return createAction(action)._cancellationType;
}