'use strict';

var _src = require('../src');

describe('listeners', function () {

      describe('success listeners', function () {

            it('should call an action success listener when action succeeds', function () {

                  var action = function action() {};
                  var actionSuccessHandler = spy();

                  (0, _src.onSuccess)(action, actionSuccessHandler);

                  (0, _src.call)(action);

                  expect(actionSuccessHandler.called).equals(true);
            });

            it('should call a global success listener when action succeeds', function () {

                  var action = function action() {};
                  var actionSuccessHandler = spy();

                  (0, _src.onSuccess)(actionSuccessHandler);

                  (0, _src.call)(action);

                  expect(actionSuccessHandler.called).equals(true);
            });

            it('should call a global success listener with action, payload and args', function () {

                  var actionFn = function actionFn() {
                        return 4;
                  };
                  var actionSuccessHandler = spy();

                  (0, _src.onSuccess)(actionSuccessHandler);

                  (0, _src.call)(actionFn, 1, 2, 3);

                  var _actionSuccessHandler = actionSuccessHandler.getCall(0).args[0],
                      payload = _actionSuccessHandler.payload,
                      args = _actionSuccessHandler.args,
                      action = _actionSuccessHandler.action;


                  expect(args).deep.equal([1, 2, 3]);
                  expect(action).equals(actionFn);
                  expect(payload).equals(4);
            });

            it('should NOT call an action success listener when action fails', function () {

                  var action = function action() {
                        throw new Error('ErrorMessage');
                  };
                  var actionSuccessHandler = spy();

                  (0, _src.onSuccess)(action, actionSuccessHandler);

                  expect(_src.call.bind(null, action)).to.throw();
                  expect(actionSuccessHandler.called).equals(false);
            });

            it('should NOT call a global success listener when action fails', function () {

                  var action = function action() {
                        throw new Error('ErrorMessage');
                  };
                  var actionSuccessHandler = spy();

                  (0, _src.onSuccess)(actionSuccessHandler);

                  expect(_src.call.bind(null, action)).to.throw();
                  expect(actionSuccessHandler.called).equals(false);
            });

            it('should NOT call an action success listener when action is cancelled', function (done) {
                  var actionSuccessHandler = spy();
                  var action = function action() {
                        return new Promise(function (resolve) {
                              return setTimeout(function () {
                                    resolve();
                                    expect(actionSuccessHandler.called).equals(false);
                                    done();
                              }, 20);
                        });
                  };

                  (0, _src.onSuccess)(action, actionSuccessHandler);
                  (0, _src.call)(action).cancel();
            });

            it('should NOT call a global success listener when action is cancelled', function (done) {
                  var actionSuccessHandler = spy();
                  var action = function action() {
                        return new Promise(function (resolve) {
                              return setTimeout(function () {
                                    resolve();
                                    expect(actionSuccessHandler.called).equals(false);
                                    done();
                              }, 20);
                        });
                  };

                  (0, _src.onSuccess)(actionSuccessHandler);
                  (0, _src.call)(action).cancel();
            });
      });

      describe('error listeners', function () {

            it('should call an action error listener when action fails', function () {
                  var action = function action() {
                        throw new Error('Error');
                  };
                  var actionErrorHandler = spy();

                  (0, _src.onError)(action, actionErrorHandler);

                  expect(_src.call.bind(null, action)).to.throw();
                  expect(actionErrorHandler.called).equals(true);
            });

            it('should call a global error listener when action fails', function () {
                  var action = function action() {
                        throw new Error('Error');
                  };
                  var actionErrorHandler = spy();

                  (0, _src.onError)(actionErrorHandler);

                  expect(_src.call.bind(null, action)).to.throw();
                  expect(actionErrorHandler.called).equals(true);
            });

            it('should call a global error listener with error and args', function () {
                  var someError = new Error('someErrorMessage');
                  var action = function action() {
                        throw someError;
                  };
                  var errorSuccessHandler = spy();

                  (0, _src.onError)(errorSuccessHandler);

                  expect(_src.call.bind(null, action, 1, 2, 3)).to.throw();

                  var _errorSuccessHandler$ = errorSuccessHandler.getCall(0).args[0],
                      args = _errorSuccessHandler$.args,
                      error = _errorSuccessHandler$.error;


                  expect(args).deep.equal([1, 2, 3]);
                  expect(error).equals(someError);
            });

            it('should NOT call an action error listener when action succeeds', function () {

                  var action = function action() {};
                  var actionErrorHandler = spy();

                  (0, _src.onError)(action, actionErrorHandler);

                  (0, _src.call)(action);
                  expect(actionErrorHandler.called).equals(false);
            });

            it('should NOT call a global error listener when action succeeds', function () {

                  var action = function action() {};
                  var actionErrorHandler = spy();

                  (0, _src.onError)(actionErrorHandler);

                  (0, _src.call)(action);

                  expect(actionErrorHandler.called).equals(false);
            });

            it('should NOT call an action error listener when action is cancelled', function (done) {
                  var actionErrorHandler = spy();
                  var action = function action() {
                        return new Promise(function (resolve) {
                              return setTimeout(function () {
                                    resolve();
                                    expect(actionErrorHandler.called).equals(false);
                                    done();
                              }, 20);
                        });
                  };

                  (0, _src.onSuccess)(action, actionErrorHandler);
                  (0, _src.call)(action).cancel();
            });

            it('should NOT call a global error listener when action is cancelled', function (done) {
                  var actionErrorHandler = spy();
                  var action = function action() {
                        return new Promise(function (resolve) {
                              return setTimeout(function () {
                                    resolve();
                                    expect(actionErrorHandler.called).equals(false);
                                    done();
                              }, 20);
                        });
                  };

                  (0, _src.onError)(actionErrorHandler);
                  (0, _src.call)(action).cancel();
            });
      });
});